from itertools import product
from typing import Callable

SIZES = [3, 5, 10, 15, 25]

def accuracy(predictions, labels):
    correct_predictions = sum(1 for prediction in predictions if prediction in labels)
    return correct_predictions / len(labels)

def calculate_recall(predictions, labels):
    correct_predictions = sum(1 for label in labels if label in predictions)
    if labels:
        return correct_predictions / len(labels)
    return 0

def calculate_reciprocal_rank(predictions, labels):
    for index, prediction in enumerate(predictions):
        if prediction in labels:
            return 1 / (index + 1)
    return 0

def calculate_precision(predictions, labels):
    correct_predictions = sum(1 for prediction in predictions if prediction in labels)
    if predictions:
        return correct_predictions / len(predictions)
    return 0

def score_retrieval(
    preds: list[str],
    label: str | list[str],
    sizes: list[int],
    scoring_fns: dict[str, Callable[[list[str], list[str]], float]],
):
    return {
        f"{fn_name}@{size}": round(
            scoring_fns[fn_name](
                preds[:size], [label] if isinstance(label, str) else label
            ),
            3,
        )
        for fn_name, size in product(scoring_fns.keys(), sizes)
    }


from datasets import load_dataset, Dataset
from tqdm import tqdm
import json

def get_labels(file_path):
    with open(file_path, 'r') as f:
        labels = []
        for line in f:
            labels.append(json.loads(line))
    return labels


import instructor
import google.generativeai as genai

class GeminiHelper():
    def __init__(self, GOOGLE_API_KEY: str):
        genai.configure(api_key=GOOGLE_API_KEY)

        self.client = instructor.from_gemini(
            client=genai.GenerativeModel(
                model_name='models/gemini-1.5-flash-latest',
            ),
            mode=instructor.Mode.GEMINI_JSON,
        )
    
    def get_client(self):
        return self.client


from lancedb.pydantic import LanceModel
from itertools import batched
from tqdm import tqdm

def get_table(db, table_name: str, schema: LanceModel | None = None):
    if table_name in db.table_names():
        return db.open_table(table_name)
    
    if schema is None:
        raise ValueError(
            f"Table {table_name} does not exist and no schema was provided"
        )
    
    return db.create_table(table_name, schema=schema, mode="overwrite")

def insert_data_into_table(table, data, batch_size=20):
    batches = batched(data, batch_size)

    for batch in tqdm(batches):
        table.add(list(batch))


import os
import logging
import torch
from itertools import batched
from tqdm import tqdm
from typing import Optional
from sentence_transformers import SentenceTransformer
from optimum.onnxruntime import ORTModelForFeatureExtraction
from transformers import AutoTokenizer
import torch.nn.functional as F
from .models import QueryItem, FormatType


logger = logging.getLogger(__name__)
class EmbeddingModel(object):
    def __init__(self, 
        model_name_or_path: str,
        format: FormatType,
        cache_dir: Optional[str] = None,
        file_name: Optional[str] = None,
    ):
        print(format)
        if format == None:
                raise ValueError("Format is required")
        else:
            logger.info("Format: {}".format(format))
            self.format_model = format

        if model_name_or_path is not None and model_name_or_path != "": 
            logger.info("Load pretrained {}: {}".format(format, model_name_or_path))
            if os.path.exists(model_name_or_path):
                model_path = model_name_or_path
            else:
                raise ValueError("Model path does not exist: {}".format(model_name_or_path))

            if format == FormatType.onnx:
                self.model = ORTModelForFeatureExtraction.from_pretrained(model_path, file_name=file_name)
                self.tokenizer = AutoTokenizer.from_pretrained(model_path)
            else: 
                self.model = SentenceTransformer(model_path)

    def _mean_pooling(self, model_output, attention_mask):
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(model_output.size()).float()
        return torch.sum(model_output * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    
    def encode(self, query: str):
        if self.format_model == FormatType.sentence_transformers:
            return self.model.encode([query])
        else:
            inputs = self.tokenizer(
                [query],
                padding=True,
                truncation=True,
                return_tensors="pt",
            )
            with torch.no_grad():
                model_outputs = self.model(**inputs)
            embeddings = self._mean_pooling(model_outputs['last_hidden_state'], inputs['attention_mask'])
            embeddings = F.normalize(embeddings, p=2, dim=1)
            return embeddings.tolist()[0]
        
    def generate_embeddings(self, data: list[QueryItem], batch_size: int):
        batches = batched(data, batch_size)

        def generate_embeddings_batch(batch):
            if self.format_model == FormatType.sentence_transformers:
                embeddings = self.model.encode(
                    [item.query for item in list(batch)]
                )
                return embeddings.tolist()
            else:
                inputs = self.tokenizer(
                    [item.query for item in list(batch)],
                    padding=True,
                    truncation=True,
                    return_tensors="pt",
                )
                with torch.no_grad():
                    model_outputs = self.model(**inputs)
                embeddings = self._mean_pooling(model_outputs['last_hidden_state'], inputs['attention_mask'])
                embeddings = F.normalize(embeddings, p=2, dim=1)
                return embeddings.tolist()
            
        batched_embeddings = [generate_embeddings_batch(batch) for batch in batches]
        
        embeddings_all = []
        for embeddings in tqdm(batched_embeddings, desc=f"Generating embeddings for {len(data)} queries using {self.format_model}"):
            for embedding in embeddings:
                embeddings_all.append(embedding)
        return embeddings_all
    





from lancedb.table import Table
from tqdm import tqdm
from lancedb.rerankers import LinearCombinationReranker, CrossEncoderReranker
import instructor
import google.generativeai as genai
from pydantic import BaseModel, Field
from typing import Literal
from tqdm.asyncio import tqdm_asyncio as asyncio
from .string_helpers import strip_punctuation
from .models import QueryItem
from .gemini_helper import GeminiHelper
from .embeddings import EmbeddingModel 
from .models import FormatType

def fts_search(table: Table, queries: list[QueryItem], top_k: int):
    data = []
    for query in tqdm(queries, desc="Excuting Full Text search now..."):
        items = (
            table.search(strip_punctuation(query.query), query_type="fts")
            .limit(top_k)
            .select(["query", "selected_chunk_ids"])
            .to_list()
        )
        data.append(items)
    return data

def vector_search(
    table: Table, queries: list[QueryItem], top_k: int, batch_size: int = 20, 
):  
    model = EmbeddingModel("data/onnx", 
                            format=FormatType.onnx, 
                            file_name='model_quantized.onnx')
    embedded_queries = model.generate_embeddings(queries, batch_size=batch_size)
    return [
        table.search(query_embedding, query_type="vector")
        .limit(top_k)
        .select(["selected_chunk_ids"])
        .to_list()
        for query_embedding in tqdm(embedded_queries, desc="Executing Vector search now...")
    ]

def hybird_search(
    table: Table, queries: list[QueryItem], top_k: int, batch_size: int = 20,
):
    return [
        table.search(strip_punctuation(query.query), query_type='hybrid')
        .limit(top_k)
        .to_list()
        for query in tqdm(queries, desc="Executing Hybrid search now...")
    ]

def linear_combination_search(
    table: Table, queries: list[QueryItem], top_k: int, vector_search_weight: float
):
    reranker = LinearCombinationReranker(vector_search_weight)
    return [
        table.search(strip_punctuation(query.query), query_type='hybrid')
        .rerank(reranker)
        .limit(top_k)
        .to_list()
        for query in tqdm(queries, desc=f"Linear Combination (weight {vector_search_weight})")
    ]

def Cross_rerank_search(
    table: Table, queries: list[QueryItem], top_k: int, model_name: str, query_type='fts',
):
    cross_reranker = CrossEncoderReranker(model_name)
    return [
        table.search(strip_punctuation(query.query), query_type=query_type)
        .rerank(cross_reranker)
        .limit(top_k)
        .to_list()
        for query in tqdm(queries, desc=f"Cross Rerank search with {model_name}")
    ]

async def classify_queries(queries: list[str], GOOGLE: str, client):
    client = client
    category_description = """
    This represents a categorization of the user's query

    - 'support': Covers queries that are asking for support or help
    - 'information': Covers queries that are asking for information
    """
    class Category(BaseModel):
        category: Literal['support', 'information'] = Field(
            ...,
            description=category_description,
        )
    async def classify(query: str):
        return await client.messages.create(
            messages= [
                {
                    "role": "system",
                    "content": "You are expert topic classifier. Your job is to classify the following title into the categories provided in the response object. Make sure to classify it into one of the individual categories provided",
                },
                {
                    "role": "user",
                    "content": f"The title is {query}",
                },
            ],
            response_model=Category,
        )
    coros = [classify(query) for query in queries]
    result = await asyncio.gather(*coros)
    return result


from lancedb.pydantic import LanceModel, Vector
from lancedb.embeddings import get_registry
from enum import Enum

class MacroModel(LanceModel):
    query: str
    selected_chunk_ids: str
    vector: Vector(1024) # type: ignore

class ViHealthModel(LanceModel):
    id: str
    query: str
    answer: str
    link: str
    vector: Vector(1024) # type: ignore

class QueryItem(LanceModel):
    query: str
    selected_chunk_ids: list[str]

class FormatType(Enum):
    sentence_transformers = "sentence-transformers"
    onnx = "onnx"

